//========================================================
// L1 solvex..
//
// Solve the (over determained) linear system Ax = b by minimized the error 
// using the  L1 metric. 
// M equations of K variables M > K
//
// The solvex uses single phase simplex algorithm to solve
// the problem:
//
// min: sum w|z|
// s.t.
// Ax + |z| = b
//
// where w is optional weight vector.
//
// This linear problem has the following proparties:
// 1. The size of the problem is MxN where N = 2(K + M).
// 2. It has an immediate basic feasible solution given by:
//    x = 0, |z| = b. 2K + M are have zero value M has non-zero value
//    which forms a basis. 
// 3. The problem has lower bound of 0 since the coeficients (w) of the objective
//    variables (z) are all positive.
// 4. (2) and (3) implies that a solution always exists and that one phase is sufficient.
//
// Double precission accuracy is used;
//
// (c) The Hebrew University of Jerusalem
// Moshe Ben-Ezra
//========================================================

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <memory.h>
#include <math.h>

#include "solvex.h"

#ifndef HUGE
#define HUGE 1e100;
#endif

#define EPSILON 1.0e-10

void solvex::reset()
{
  cur_line = 1;
  memset(tab,0,lpwidth * lpheight * sizeof(double));
  memset(B,0,lpheight * sizeof(int));
  memset(sol,0,lpwidth * sizeof(double));
}


// m equations, k variables

solvex::solvex(int m, int k)
{
  equ_no = m;
  var_no = k;
  
  lpvar_no = 2*(var_no + equ_no); // 2 * original vars + slack vars  
  lpwidth  = lpvar_no + 1;	  // number of variables + b row
  lpheight = equ_no + 1;	  // number of equations + obective function row
  
  
  tab = new double[lpwidth * lpheight];
  assert(tab != NULL);
  
  sol = new double[lpwidth];
  assert(tab != NULL);
  
  B = new int[lpheight];	// first entry unused.
  assert(B != NULL);
  
  reset();
}


solvex::~solvex()
{
  delete [] tab;
  delete [] B;
  delete [] sol;
}

double *solvex::get_row(int i) {
  return tab + i*lpwidth;
}


// need to unroll
void solvex::negate_row(double *p)
{
  int i;
  for (i=0; i < lpwidth; i++, p++) *p = -(*p);
}

// need to unroll
void solvex::multiply_row(double *p, double scalar)
{
  int i;
  for (i=0; i < lpwidth; i++, p++) *p = scalar * (*p);
}

// This function takes 96.7% of the time. Optimize it !
// the scalar is usually != 1 so it doesn't help to check this


void solvex::subtruct_row(double *src, double *dst, double scalar = 1.0)
{
  register int i;
  
  register int blocks = lpwidth / 12;
  register int lefts   = lpwidth % 12;
  
  for (i=0; i < blocks; i++, src+=12, dst+=12){
    *dst      -= scalar * (*src);
    *(dst+1)  -= scalar * (*(src+1));
    *(dst+2)  -= scalar * (*(src+2));
    *(dst+3)  -= scalar * (*(src+3));
    *(dst+4)  -= scalar * (*(src+4));
    *(dst+5)  -= scalar * (*(src+5));
    *(dst+6)  -= scalar * (*(src+6));
    *(dst+7)  -= scalar * (*(src+7));
    *(dst+8)  -= scalar * (*(src+8));
    *(dst+9)  -= scalar * (*(src+9));
    *(dst+10) -= scalar * (*(src+10));
    *(dst+11) -= scalar * (*(src+11));
  }
  
  for (i=0; i < lefts; i++, src++, dst++){
    *dst -= scalar * (*src);
  }
}


int solvex::add_equ(double *p, double val, double w)
{
  double *line, *obj;
  int i, slack_offset;
  
  if (cur_line >= lpheight) return -1;
  
  line = get_row(cur_line);	// first line is the objective function.
  slack_offset = 2*var_no + 2*(cur_line-1)+1;	// Z location
  
  line[0] = val;		// b vector
  for (i=0; i<var_no; i++){	// A matrix
    line[2*i+1] = +p[i];
    line[2*i+2] = -p[i];
  }
  line[slack_offset]   = +1;	// +Zi
  line[slack_offset+1] = -1;	// -Zi
  
  obj = get_row(0);		
  obj[slack_offset]   = w;	// Objective function 
  obj[slack_offset+1] = w;	// Objective function
  
  if (val >= 0){								
    B[cur_line] = slack_offset;	// index for feasible solution.
  }
  else{
    B[cur_line] = slack_offset + 1 ; // index for feasible solution.	
    negate_row(line);		// set to one by row operation.
  }
  
  cur_line ++;
  
  return 0;
}

int solvex::locate_pivot(int &column, int &row, double &pval)
{
  int i;
  double min, ratio, *line, tval;
  
  // Heuristic column the most negative Ci member
  // note that basic solution have now Ci = 0
  // so the are not included.
  min = 0;
  for (i=1; i < lpwidth; i++){
    if (tab[i] < min){
      min = tab[i];
      column = i;
    }
  }
  
  if (min >= -EPSILON)			// no more refinment possible
    return RET_OK;
  
  // Find best cost row, pivot even if the row is degenerated.
  
  min = HUGE;			// this allow having some value for
  // degnerated problem.
  int okflag;
  
  okflag = 0;
  
  for (i=1; i < lpheight; i++){
    line = get_row(i);
    tval = line[column];
    
    if (tval < EPSILON)	
      continue;			
    
    okflag = 1;
    
    ratio = line[0] / tval;
    if (ratio < min){
      min = ratio;
      row = i;
      pval = tval;
    }
  }
  
  if (okflag)
    return RET_CONT;
  else
    return RET_DIVZ; 
}

void solvex::pivot(int col,int row, double pval)
{
  int i;
  double *line, *line2, scalar;
  
  line = get_row(row);
  multiply_row(line,1/pval);
  
  for (i=0; i< lpheight; i++){
    if (i == row) continue;
    line2 = get_row(i);
    scalar = line2[col];
    if (fabs(scalar) > EPSILON)
	subtruct_row(line,line2,scalar); // zero A[i,col]
  }
  
  // update B vector
  B[row] = col;
}

int solvex::solve()
{
  double *line, scalar,pval;
  int i, column, row;
  
  
  // Step 1: convert the basic feasible solution into identity basis
  // we almost there - we just need to zero the objective coeficients for
  // the basic solution columns.
  
  for (i=1; i<=equ_no; i++){
    line = get_row(i);
    scalar = tab[B[i]];		// objective function coeficient
    subtruct_row(line,tab,scalar); // zero objective function coeficient. 
  }
  
  // Step 2: pivoting
  int ret_val;
  while((ret_val = locate_pivot(column, row,pval)) == RET_CONT){
    pivot(column,row,pval);
    if (-tab[0] < -EPSILON) return RET_FAIL;		// Objective function cannot be negative !
  } 
  
  if (ret_val == RET_DIVZ)
    return RET_DIVZ;
  
  // Step 3: construct solution (patch non zero bfs into sol
  
  for (i=1; i<lpheight; i++){
    sol[B[i]] = get_row(i)[0];
  }
  
  return RET_OK;
}

double solvex::get_obj()
{
  return -(*tab);
}

double solvex::get_var(int i)
{
  return sol[2*i+1] - sol[2*i+2];
}



void solvex::print(FILE *fp)
{
  for (int i=0; i<lpheight; i++){
    for (int j=0; j<lpwidth; j++){
      fprintf(fp,"%7.3f ",get_row(i)[j]);
    }
    putc('\n',fp);
  }
}


#ifdef TEST

//----------------------------------------------------------------
// The following code tests the solver using a simple 3x3 problem, 
// It also demonstrates its usage
//----------------------------------------------------------------

main()
{
  double e1[] = {2, -5, 9};
  double e2[] = {7, 12, 3};
  double e3[] = {11, -4, 6};

  double X1, X2, X3;
  
  // Defind problem size 3 eqations and 3 variables
  solvex lp(3,3);

  // add equations 
  lp.add_equ(e1,18); // e1(1)X1 + e1(2)X2 + e1(3)X3 = 18
  lp.add_equ(e2,-6);
  lp.add_equ(e3,2);

  // Solve the system, the return value should be zero is no error occurred
  printf("%d\n",lp.solve());
  
  // Print the objective function (the L1 minimum error)
  printf("Obj: %g\n",lp.get_obj());

  // These are the solutions
  printf("X0 = %g\n",X1=lp.get_var(0));
  printf("X1 = %g\n",X2=lp.get_var(1));
  printf("X2 = %g\n",X3=lp.get_var(2));
  
  // These are the error values for each equation
  printf("E1 = %g\n",lp.get_var(3));
  printf("E2 = %g\n",lp.get_var(4));
  printf("E3 = %g\n",lp.get_var(5));

  // Verify the solution
  printf("Verify1: %g\n", e1[0]*X1 + e1[1]*X2 +e1[2]*X3 - 18);
  printf("Verify2: %g\n", e2[0]*X1 + e2[1]*X2 +e2[2]*X3  + 6);
  printf("Verify3: %g\n", e3[0]*X1 + e3[1]*X2 +e3[2]*X3 - 2);

  return 0;
}

#endif
